/* -*- mode: C; tab-width:8;  -*-

             fxvsetup.c - 3Dfx VooDoo vertices setup functions 
*/

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file fxapi.c for more informations about authors
 *
 */

#if defined(FX)

#include "fxdrv.h"

#define GOURAUD_ENABLED  0x01
#define TEXTURE0_ENABLED 0x02
#define FOG_ENABLED      0x04
#define ZBUFFER_ENABLED  0x08
#define TEXTURE1_ENABLED 0x10

/************************************************************************/
/********* The generic Version of the asm GrVertex Setup code ***********/
/************************************************************************/

#define NOSNAP(v) { \
  (v)->x=VB->Win[i][0];	         \
  (v)->y=VB->Win[i][1];	         \
}

#define SNAP(v) { \
  /* trunc (x,y) to multiple of 1/16 */		        \
  (v)->x=((int)(VB->Win[i][0]*16.0f))*(1.0f/16.0f);	\
  (v)->y=((int)(VB->Win[i][1]*16.0f))*(1.0f/16.0f);	\
}


#define GOURAUD(v) { \
  (v)->r=(float) (VB->Color[i][0]); \
  (v)->g=(float) (VB->Color[i][1]); \
  (v)->b=(float) (VB->Color[i][2]); \
  (v)->a=(float) (VB->Color[i][3]); \
}

#define TEXTURE0_TMU0(v)  { \
  (v)->tmuvtx[0].sow=sscale0*VBtcoord0[i][0]*(v)->oow; \
  (v)->tmuvtx[0].tow=tscale0*VBtcoord0[i][1]*(v)->oow; \
  (v)->tmuvtx[0].oow=VBtcoord0[i][3]*(v)->oow; \
}

#define TEXTURE0_TMU1(v)  { \
  (v)->tmuvtx[1].sow=sscale0*VBtcoord0[i][0]*(v)->oow/VBtcoord0[i][3]; \
  (v)->tmuvtx[1].tow=tscale0*VBtcoord0[i][1]*(v)->oow/VBtcoord0[i][3]; \
  (v)->tmuvtx[1].oow=VBtcoord0[i][3]*(v)->oow; \
}

#define TEXTURE1_TMU1(v)  { \
  (v)->tmuvtx[1].sow=sscale1*VBtcoord1[i][0]*(v)->oow/VBtcoord1[i][3]; \
  (v)->tmuvtx[1].tow=tscale1*VBtcoord1[i][1]*(v)->oow/VBtcoord1[i][3]; \
  (v)->tmuvtx[1].oow=VBtcoord1[i][3]*(v)->oow; \
}

#define TEXTURE1_TMU0(v)  { \
  (v)->tmuvtx[0].sow=sscale1*VBtcoord1[i][0]*(v)->oow/VBtcoord1[i][3]; \
  (v)->tmuvtx[0].tow=tscale1*VBtcoord1[i][1]*(v)->oow/VBtcoord1[i][3]; \
  (v)->tmuvtx[0].oow=VBtcoord1[i][3]*(v)->oow; \
}

#define WBUFFER(v) { (v)->oow=wscale/VB->Clip[i][3]; }

#define ZBUFFER(v) { (v)->ooz=VB->Win[i][2]; }

#define NOP(v) 

#define FXSETUP(label, snap, gouraud, texture0, texture1, wdepth, zdepth) {    \
  register unsigned int i;					        \
  register struct vertex_buffer *VB = ctx->VB;			        \
  register GLfloat (*VBtcoord0)[4];                                     \
  register GLfloat (*VBtcoord1)[4];                                     \
  fxMesaContext fxMesa = (fxMesaContext)ctx->DriverCtx; 		\
  register GrVertex *GVB = &fxMesa->gWin[vstart];			\
  register float wscale = fxMesa->wScale; 			        \
  register float sscale0;                                               \
  register float tscale0;                                               \
  register float sscale1;                                               \
  register float tscale1;                                               \
								        \
  if(ctx->Texture.Unit[0].Current && ctx->Texture.Unit[0].Current->DriverData) { \
    sscale0=((tfxTexInfo *)(ctx->Texture.Unit[0].Current->DriverData))->sScale; \
    tscale0=((tfxTexInfo *)(ctx->Texture.Unit[0].Current->DriverData))->tScale; \
    VBtcoord0=VB->MultiTexCoord[0];              \
  }                                                                            \
  if(ctx->Texture.Unit[1].Current && ctx->Texture.Unit[1].Current->DriverData) { \
    sscale1=((tfxTexInfo *)(ctx->Texture.Unit[1].Current->DriverData))->sScale; \
    tscale1=((tfxTexInfo *)(ctx->Texture.Unit[1].Current->DriverData))->tScale; \
    VBtcoord1=VB->MultiTexCoord[1];              \
  }                                                           \
                                                              \
  for(i=vstart;i<vend;i++,GVB++) {                            \
    if(VB->ClipMask[i]==0) {                                  \
      snap(GVB);                                              \
      gouraud(GVB);                                           \
      wdepth(GVB);                                            \
      zdepth(GVB);                                            \
      texture0(GVB);                                          \
      texture1(GVB);                                          \
   }                                                          \
 }                                                            \
}

#ifdef FX_V2
#define FXVSETUP_FUNC(fname, gouraud, texture0, texture1, wdepth, zdepth) \
static void fname (GLcontext *ctx, GLuint vstart, GLuint vend)        \
{                                                                     \
  FXSETUP(fname##_l1,NOSNAP,gouraud,texture0,texture1,wdepth,zdepth); \
}
#else
#define FXVSETUP_FUNC(fname, gouraud, texture0, texture1, wdepth, zdepth) \
static void fname (GLcontext *ctx, GLuint vstart, GLuint vend)        \
{                                                                     \
  if(ctx->Primitive==GL_POINTS) {                                     \
    FXSETUP(fname##_l1,NOSNAP,gouraud,texture0,texture1,wdepth,zdepth);   \
  } else {                                                            \
    FXSETUP(fname##_l2,SNAP,gouraud,texture0,texture1,wdepth,zdepth); \
  }                                                                   \
}
#endif

/********************************************************/
/************************T00T11**************************/
/********************************************************/

FXVSETUP_FUNC(fxSetup,NOP,NOP,NOP,NOP,NOP)
FXVSETUP_FUNC(fxSetupG,GOURAUD,NOP,NOP,NOP,NOP)
FXVSETUP_FUNC(fxSetupT00,NOP,TEXTURE0_TMU0,NOP,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupGT00,GOURAUD,TEXTURE0_TMU0,NOP,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupW,NOP,NOP,NOP,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupGW,GOURAUD,NOP,NOP,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupZ,NOP,NOP,NOP,NOP,ZBUFFER)
FXVSETUP_FUNC(fxSetupWZ,NOP,NOP,NOP,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupGZ,GOURAUD,NOP,NOP,NOP,ZBUFFER)
FXVSETUP_FUNC(fxSetupGWZ,GOURAUD,NOP,NOP,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupT00Z,NOP,TEXTURE0_TMU0,NOP,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupGT00Z,GOURAUD,TEXTURE0_TMU0,NOP,WBUFFER,ZBUFFER)

/*** FxSetup functions for Multi-texture ***/

FXVSETUP_FUNC(fxSetupT11,NOP,NOP,TEXTURE1_TMU1,NOP,NOP)
FXVSETUP_FUNC(fxSetupGT11,GOURAUD,NOP,TEXTURE1_TMU1,NOP,NOP)
FXVSETUP_FUNC(fxSetupT00T11,NOP,TEXTURE0_TMU0,TEXTURE1_TMU1,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupGT00T11,GOURAUD,TEXTURE0_TMU0,TEXTURE1_TMU1,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupWT11,NOP,NOP,TEXTURE1_TMU1,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupGWT11,GOURAUD,NOP,TEXTURE1_TMU1,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupZT11,NOP,NOP,TEXTURE1_TMU1,NOP,ZBUFFER)
FXVSETUP_FUNC(fxSetupWZT11,NOP,NOP,TEXTURE1_TMU1,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupGZT11,GOURAUD,NOP,TEXTURE1_TMU1,NOP,ZBUFFER)
FXVSETUP_FUNC(fxSetupGWZT11,GOURAUD,NOP,TEXTURE1_TMU1,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupT00ZT11,NOP,TEXTURE0_TMU0,TEXTURE1_TMU1,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupGT00ZT11,GOURAUD,TEXTURE0_TMU0,TEXTURE1_TMU1,WBUFFER,ZBUFFER)

/********************************************************/
/************************T01T10**************************/
/********************************************************/

FXVSETUP_FUNC(fxSetupT01,NOP,TEXTURE0_TMU1,NOP,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupGT01,GOURAUD,TEXTURE0_TMU1,NOP,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupT01Z,NOP,TEXTURE0_TMU1,NOP,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupGT01Z,GOURAUD,TEXTURE0_TMU1,NOP,WBUFFER,ZBUFFER)

/*** FxSetup functions for Multi-texture ***/

FXVSETUP_FUNC(fxSetupT10,NOP,NOP,TEXTURE1_TMU0,NOP,NOP)
FXVSETUP_FUNC(fxSetupGT10,GOURAUD,NOP,TEXTURE1_TMU0,NOP,NOP)
FXVSETUP_FUNC(fxSetupT01T10,NOP,TEXTURE0_TMU1,TEXTURE1_TMU0,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupGT01T10,GOURAUD,TEXTURE0_TMU1,TEXTURE1_TMU0,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupWT10,NOP,NOP,TEXTURE1_TMU0,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupGWT10,GOURAUD,NOP,TEXTURE1_TMU0,WBUFFER,NOP)
FXVSETUP_FUNC(fxSetupZT10,NOP,NOP,TEXTURE1_TMU0,NOP,ZBUFFER)
FXVSETUP_FUNC(fxSetupWZT10,NOP,NOP,TEXTURE1_TMU0,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupGZT10,GOURAUD,NOP,TEXTURE1_TMU0,NOP,ZBUFFER)
FXVSETUP_FUNC(fxSetupGWZT10,GOURAUD,NOP,TEXTURE1_TMU0,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupT01ZT10,NOP,TEXTURE0_TMU1,TEXTURE1_TMU0,WBUFFER,ZBUFFER)
FXVSETUP_FUNC(fxSetupGT01ZT10,GOURAUD,TEXTURE0_TMU1,TEXTURE1_TMU0,WBUFFER,ZBUFFER)

/********************************************************/

static tfxSetupFunc fxSetupFuncsT00T11[]={
  fxSetup,
  fxSetupG,
  fxSetupT00,
  fxSetupGT00,
  fxSetupW, /* 4 */
  fxSetupGW,
  fxSetupT00,
  fxSetupGT00,
  fxSetupZ, /* 8 */
  fxSetupGZ,
  fxSetupT00Z,
  fxSetupGT00Z,
  fxSetupWZ, /* 12 */
  fxSetupGWZ,
  fxSetupT00Z,
  fxSetupGT00Z,
  fxSetupT11, /* 16 */
  fxSetupGT11,
  fxSetupT00T11,
  fxSetupGT00T11,
  fxSetupWT11, /* 20 */
  fxSetupGWT11,
  fxSetupT00T11,
  fxSetupGT00T11,
  fxSetupZT11, /* 24 */
  fxSetupGZT11,
  fxSetupT00ZT11,
  fxSetupGT00ZT11,
  fxSetupWZT11, /* 28 */
  fxSetupGWZT11,
  fxSetupT00ZT11,
  fxSetupGT00ZT11
};

static tfxSetupFunc fxSetupFuncsT01T10[]={
  fxSetup,
  fxSetupG,
  fxSetupT01,
  fxSetupGT01,
  fxSetupW, /* 4 */
  fxSetupGW,
  fxSetupT01,
  fxSetupGT01,
  fxSetupZ, /* 8 */
  fxSetupGZ,
  fxSetupT01Z,
  fxSetupGT01Z,
  fxSetupWZ, /* 12 */
  fxSetupGWZ,
  fxSetupT01Z,
  fxSetupGT01Z,
  fxSetupT10, /* 16 */
  fxSetupGT10,
  fxSetupT01T10,
  fxSetupGT01T10,
  fxSetupWT10, /* 20 */
  fxSetupGWT10,
  fxSetupT01T10,
  fxSetupGT01T10,
  fxSetupZT10, /* 24 */
  fxSetupGZT10,
  fxSetupT01ZT10,
  fxSetupGT01ZT10,
  fxSetupWZT10, /* 28 */
  fxSetupGWZT10,
  fxSetupT01ZT10,
  fxSetupGT01ZT10
};

tfxSetupFunc fxDDChooseSetupFunction(GLcontext *ctx)
{
  tfxSetupFunc *fxSetupFuncs=fxSetupFuncsT00T11;
  unsigned int setupindex=0;

  grHints(GR_HINT_STWHINT,GR_STWHINT_W_DIFF_TMU0);

  if(ctx->Light.ShadeModel == GL_SMOOTH && !ctx->Light.Model.TwoSide)
    setupindex |= GOURAUD_ENABLED;

  if((ctx->Texture.Enabled & TEXTURE0_2D) &&
     !(ctx->Texture.Enabled & TEXTURE0_3D) &&
     (ctx->Texture.Unit[0].Current2D->Complete))
    setupindex |= TEXTURE0_ENABLED;

  if((ctx->Texture.Enabled & TEXTURE1_2D) &&
     !(ctx->Texture.Enabled & TEXTURE1_3D) &&
     (ctx->Texture.Unit[1].Current2D->Complete)) {
    setupindex |= TEXTURE1_ENABLED;

    if(!(setupindex & TEXTURE0_ENABLED))
      fxSetupFuncs=fxSetupFuncsT01T10;
    else {
      struct gl_texture_object *tObj=ctx->Texture.Unit[0].Current2D;
      tfxTexInfo *ti=(tfxTexInfo *)tObj->DriverData;

      if(ti->tmi.whichTMU!=FX_TMU0)
	fxSetupFuncs=fxSetupFuncsT01T10;
    }
  }

  if(ctx->Fog.Enabled && (ctx->FogMode==FOG_FRAGMENT))
    setupindex |= FOG_ENABLED;

  if(ctx->Depth.Test)
    setupindex |= ZBUFFER_ENABLED;

#if defined(DEBUG_FXMESA)
    fprintf(stderr,"fxmesa: vertex setup function %d (%d,%d,%d)\n",
	    setupindex,setupindex & GOURAUD_ENABLED, setupindex & TEXTURE0_ENABLED, 
	    setupindex & TEXTURE1_ENABLED);
#endif

  return fxSetupFuncs[setupindex];
}

#else


/*
 * Need this to provide at least one external definition.
 */

int gl_fx_dummy_function_vsetup(void)
{
  return 0;
}

#endif  /* FX */
