/* $Id: xform.h,v 3.4 1998/11/07 14:21:55 brianp Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * Copyright (C) 1995-1998  Brian Paul
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * $Log: xform.h,v $
 * Revision 3.4  1998/11/07 14:21:55  brianp
 * replaced static array initialization with gl_init_transformation()
 *
 * Revision 3.3  1998/10/29 03:57:11  brianp
 * misc clean-up of new vertex transformation code
 *
 * Revision 3.2  1998/10/29 02:28:13  brianp
 * incorporated Keith Whitwell's transformation optimizations
 *
 * Revision 3.1  1998/02/01 16:37:19  brianp
 * added GL_EXT_rescale_normal extension
 *
 * Revision 3.0  1998/01/31 21:08:31  brianp
 * initial rev
 *
 */


#ifndef XFORM_H
#define XFORM_H


#include "types.h"


/*
 * Transform a point (column vector) by a matrix:   Q = M * P
 */
#define TRANSFORM_POINT( Q, M, P )					\
   Q[0] = M[0] * P[0] + M[4] * P[1] + M[8] *  P[2] + M[12] * P[3];	\
   Q[1] = M[1] * P[0] + M[5] * P[1] + M[9] *  P[2] + M[13] * P[3];	\
   Q[2] = M[2] * P[0] + M[6] * P[1] + M[10] * P[2] + M[14] * P[3];	\
   Q[3] = M[3] * P[0] + M[7] * P[1] + M[11] * P[2] + M[15] * P[3];


#define TRANSFORM_POINT3( Q, M, P )				\
   Q[0] = M[0] * P[0] + M[4] * P[1] + M[8] *  P[2] + M[12];	\
   Q[1] = M[1] * P[0] + M[5] * P[1] + M[9] *  P[2] + M[13];	\
   Q[2] = M[2] * P[0] + M[6] * P[1] + M[10] * P[2] + M[14];	\
   Q[3] = M[3] * P[0] + M[7] * P[1] + M[11] * P[2] + M[15];


/*
 * Transform a normal (row vector) by a matrix:  [NX NY NZ] = N * MAT
 */
#define TRANSFORM_NORMAL( TO, N, MAT )				\
do {								\
   TO[0] = N[0] * MAT[0] + N[1] * MAT[1] + N[2] * MAT[2];	\
   TO[1] = N[0] * MAT[4] + N[1] * MAT[5] + N[2] * MAT[6];	\
   TO[2] = N[0] * MAT[8] + N[1] * MAT[9] + N[2] * MAT[10];	\
} while (0)


extern void gl_transform_vector( GLfloat u[4],
				 const GLfloat v[4],
                                 const GLfloat m[16] );


extern void gl_transform_normals_3fv( GLuint n, 
				      const GLfloat *in,
				      GLuint in_stride,
				      const GLfloat m[16],
				      GLfloat out[][3], 
				      GLboolean normalize,
				      GLboolean rescale );


extern void gl_normalize_3fv( GLuint n, 
			      const GLfloat *in,
			      GLuint in_stride,
			      GLfloat out[][3] );


extern void gl_scale_3fv( GLuint n, 
			  const GLfloat *in,
			  GLuint in_stride,
			  GLfloat out[][3],
			  GLfloat scale );


extern void gl_init_transformation( void );


extern void gl_project_and_cliptest_points( const GLmatrix *mat,
					    GLuint n, 
					    const GLfloat *from,
					    GLfloat to[][4],
					    GLuint stride,
					    GLubyte clipMask[],
					    GLubyte *orMask, 
					    GLubyte *andMask,
					    GLuint vec_size );


extern void gl_transform_points( const GLmatrix *mat,
				 GLuint n,
				 const GLfloat *from, 
				 GLfloat to[][4],
				 GLuint stride,
				 GLuint vec_size );


#endif
